use std::path::PathBuf;
use std::sync::mpsc;
use std::time::Duration;

use anyhow::Result;
use crossterm::event::{self, Event, KeyCode, KeyEvent};
use ratatui::layout::Constraint;
use ratatui::style::{Color, Style};
use ratatui::widgets::{Block, Borders, Cell, Row, Table, TableState};
use ratatui::Frame;

use crate::Projects;


#[derive(Debug, Clone, PartialEq, Eq)]
enum Message {
    Exit,
    Noop,
    MoveDown,
    MoveUp,
    Confirm,
}


struct State {
    projects: Projects,
    should_exit: bool,
    project_table: TableState,
    selected_project: Option<PathBuf>,
}

impl State {
    pub fn new(projects: Projects) -> Self {
        let project_table = TableState::new().with_selected(Some(0));

        Self {
            projects,
            should_exit: false,
            project_table,
            selected_project: None,
        }
    }
}


pub fn run(projects: Projects) -> Result<Option<PathBuf>> {
    let mut term = ratatui::init();
    let mut state = State::new(projects);

    let (mut msg_tx, mut msg_rx) = mpsc::channel();

    loop {
        handle_messages(&mut state, &mut msg_rx)?;

        if state.should_exit {
            break;
        }

        term.draw(|frame| draw(&mut state, frame))?;

        if !event::poll(Duration::from_millis(10))? {
            continue;
        }

        if let Event::Key(key_event) = event::read()? {
            handle_event(&mut msg_tx, key_event);
        }
    }

    ratatui::restore();

    Ok(state.selected_project)
}


fn handle_event(tx: &mut mpsc::Sender<Message>, event: KeyEvent) {
    let msg = match event.code {
        KeyCode::Char('q') | KeyCode::Esc => Message::Exit,
        KeyCode::Char(' ') | KeyCode::Enter => Message::Confirm,
        KeyCode::Char('j') | KeyCode::Down => Message::MoveDown,
        KeyCode::Char('k') | KeyCode::Up => Message::MoveUp,
        _ => Message::Noop,
    };

    tx.send(msg).unwrap();
}


fn handle_messages(state: &mut State, rx: &mut mpsc::Receiver<Message>) -> Result<()> {
    let msg = match rx.try_recv() {
        Ok(msg) => msg,
        Err(mpsc::TryRecvError::Empty) => {
            return Ok(());
        }
        Err(mpsc::TryRecvError::Disconnected) => {
            state.should_exit = true;
            return Ok(());
        }
    };

    match msg {
        Message::Exit => {
            state.should_exit = true;
            state.selected_project = None;
        }
        Message::MoveDown => {
            state.project_table.select_next();
        }
        Message::MoveUp => {
            state.project_table.select_previous();
        }
        Message::Confirm => {
            if let Some(selected) = state.project_table.selected() {
                if let Some(project_path) = state.projects.list.get(selected) {
                    state.should_exit = true;
                    state.selected_project = Some(project_path.clone());
                };
            }
        }
        _ => (),
    }

    Ok(())
}


fn draw(state: &mut State, frame: &mut Frame) {
    let block = Block::default().borders(Borders::ALL);

    let rows = state.projects.list.iter().map(|project| {
        let project_path = project.to_str().unwrap();
        Row::new([Cell::new(project_path)])
    });

    let widths = [Constraint::Min(20)];

    let table_highlight = Style::default().fg(Color::Blue);

    let table = Table::new(rows, widths)
        .block(block)
        .row_highlight_style(table_highlight);

    frame.render_stateful_widget(table, frame.area(), &mut state.project_table);
}
